from discord.ui import Modal, InputText
from discord import InputTextStyle, Interaction
from typing import Optional, Callable, Union
from functools import partial
from asyncio import iscoroutinefunction

from .errors import CoroutineError, CustomIDNotFound

class EasyModal(Modal):

    def __init__(self, title: str, custom_id: Optional[str] = None, timeout: Optional[float] = None, global_callable: Optional[Callable] = None):
        """
        Init EasyModal instance.
        :param title: The title set to the Modal
        :param custom_id: ID set to the Modal
        :param timeout: Timeout before the view don't allow response.
        :param global_callable: Asynchronous function called when user submit the modal. Called before all subcoroutine function set for each inputText. Get the instance modal and user interaction in parameters
        """
        super().__init__(title=title, custom_id=custom_id, timeout=timeout)

        self.__callback: dict[str, Union[Callable, None]] = {}

        if global_callable is not None and not iscoroutinefunction(global_callable):
            raise CoroutineError(global_callable)

        self.__global_callable: Optional[Callable] = global_callable

    def add_input_text(self, label: str,
                       style: InputTextStyle = InputTextStyle.short,
                       custom_id: Optional[str] = None,
                       placeholder: Optional[str] = None,
                       min_length: Optional[int] = None,
                       max_length: Optional[int] = None,
                       required: Optional[bool] = True,
                       value: Optional[str] = None,
                       row: Optional[int] = None) -> Callable:
        """
        Add an input text on the Modal.
        :return: set_inputText_callable function to set the callable on the inputText. Require an asynchronous function in parameters.
        """

        x = InputText(label=label, style=style, custom_id=custom_id, placeholder=placeholder, min_length=min_length, max_length=max_length, required=required, row=row, value=value)
        self.__callback[x.custom_id] = None
        self.add_item(x)
        return partial(self.set_inputText_callable, x.custom_id)

    def set_inputText_callable(self, inputText_id: str, _callable: Callable) -> "EasyModal":
        """
        Set an asynchronous function to a single inputText. _callable function was called when modal is completed by the user

        x = EasyModal(title="test")
        # init '8878' custom_id for one inputtext

        x.set_inputText_callable(inputText_id='8878', _callable=set_any)

        async def set_any(data: InputText, interaction: Interaction):
                ...
        """
        if not iscoroutinefunction(_callable):
            raise CoroutineError(_callable)

        if inputText_id not in self.__callback.keys():
            raise CustomIDNotFound(inputText_id)

        self.__callback[inputText_id] = _callable
        return self

    def del_inputText_callable(self, inputText_id: str) -> "EasyModal":
        """
        Delete callable object link to the inputText ID
        :param inputText_id: ID to the inputText
        """
        if inputText_id not in self.__callback.keys():
            raise CustomIDNotFound(inputText_id)

        self.__callback[inputText_id] = None

    async def callback(self, interaction: Interaction):
        """
        Call when the user submit the modal.
        All callable function set to the user get the 'interaction' parameter.
        """

        if self.__global_callable is not None:
            await self.__global_callable(self, interaction)

        for inputTextID, _callable in self.__callback.items():
            if _callable is not None:
                await _callable(self.get_input_text(inputTextID),interaction)



    def get_input_text(self, inputText_id: str) -> InputText:
        """
        Return the inputText associated to the inputText_id
        :param inputText_id: ID inputText
        """
        for i in self.children:
            if i.custom_id == inputText_id:
                return i

        raise CustomIDNotFound(inputText_id)



from discord.components import ComponentType
from discord.ui import Select, Item
from discord import MISSING, Emoji, PartialEmoji, Role
from typing import Callable, Union, Optional, Any

from .errors import NotCoroutineError, ComponentTypeError


class Menu:

    def __init__(self, menu_type: ComponentType, selectmenu: "SelectMenu", **kwargs):
        """
        A basic menu from selectMenu class
        """
        self.__menu: CustomSelect = CustomSelect(menu=self, select_type=menu_type, **kwargs)
        self.__selectMenu = selectmenu
        self.__menu_type: ComponentType = menu_type

    def set_callable(self, _callable: Union[Callable, None],
                     data: Optional[dict[str, Any]] = None,
                     autorised_roles : Optional[list[Union[int, Role]]] = None,
                     autorised_key: Optional[Callable] = None) -> "Menu":
        """
        Add a coroutine to the menu (same function on SelectMenu class)
        This coroutine can have 3 parameters (X, interaction, data)
        """
        if not isinstance(_callable, Callable):
            raise NotCoroutineError(_callable)

        self.__selectMenu.set_callable(self.__menu.custom_id, _callable=_callable, data=data, autorised_roles=autorised_roles, autorised_key=autorised_key)
        return self

    def add_option(self, label: str, value: str = MISSING, description: Union[str, None] = None, emoji: Union[str, Emoji, PartialEmoji, None] = None, default: bool = False) -> "Menu":
        """
        Add an option to choice.
        Only from string_select type !
        """
        self.__is_string_select()

        self.__menu.add_option(label=label, value=value, description=description, emoji=emoji, default=default)
        return self

    def remove_options(self, *labels: str) -> "Menu":
        """
        Remove an option to choice.
        Only from string_select type !
        :param labels: Label option name to delete
        """
        self.__is_string_select()

        self.__menu.options = [i for i in self.__menu.options if i.label not in labels]
        return self

    def update_option(self, current_label: str, new_label: str = None, value: str = None, description: Union[str, None] = None, emoji: Union[str, Emoji, PartialEmoji, None] = None, default: Union[bool, None] = None) -> "Menu":
        """
        Update option. To find the option, write her actual label in "current_label" option.
        Only from string_select type !
        :param current_label: The current label option to edit
        """
        self.__is_string_select()

        for option in self.__menu.options:
            if option.label == current_label:
                option.label = new_label if new_label is not None else option.label
                option.value = value if value is not None else option.value
                option.description = description if description is not None else option.description
                option.default = default if default is not None else option.default
                option.emoji = emoji if emoji is not None else option.emoji
                break
        return self

    def __is_string_select(self) -> None:
        """
        Check if the menu is a string_select
        :raise: ComponentTypeError
        """
        if self.__menu.type != ComponentType.string_select:
            raise ComponentTypeError()

    @property
    def component(self) -> "CustomSelect":
        """
        Get the component
        """
        return self.__menu

    @property
    def selectmenu(self) -> "SelectMenu":
        """
        Get the selectMenu
        """
        return self.__selectMenu

    @property
    def callable(self) -> Callable:
        """
        Get the current callable menu
        """
        return self.__selectMenu.get_callable(self.__menu.custom_id)

class CustomSelect(Select):
    """
    Subclass of Select discord Class to use some SelectMenu functions
    """

    def __init__(self, menu: Menu, select_type: ComponentType, *items: Item, **kwargs):
        super().__init__(select_type=select_type, *items, **kwargs)
        self.__menu: Menu = menu

    async def update(self):
        """
        Bridge to SelectMenu update function
        """
        return await self.__menu.selectmenu.update()

    @property
    def get_view(self) -> "EasyModifiedViews":
        """
        Bridge to SelectMenu get_view property
        """
        return self.__menu.selectmenu.get_view



from threading import Thread
from discord import Intents, Bot, ApplicationCommand
from asyncio import run_coroutine_threadsafe, new_event_loop, set_event_loop, Future, AbstractEventLoop, sleep
from time import sleep as tsleep
from .errors import BotNotStartedError, SetupCommandFunctionNotFound, CommandFileNotFoundError
from typing import Optional
from importlib import reload
from importlib.util import spec_from_file_location, module_from_spec
from os import path
from sys import modules


class DiscordBot:

    def __init__(self, token: str, intents: Intents,  command_prefix: Optional[str] = None):
        self.__token: str = token
        self.__command_prefix: str = command_prefix
        self.__running_bot: Future = None
        self.__loop: AbstractEventLoop = new_event_loop()
        self.__thread: Thread = Thread(target=self.run_loop, daemon=True)  # Thread pour exécuter l'event loop
        self.__thread.start()
        self.__bot: Bot = Bot(loop=self.__loop, intents=intents, command_prefix=command_prefix, help_commad=None, auto_sync_commands=False)
        self.__intents: Intents = intents
        self.__imported_module: list[dict[str, ...]] = []

    def run_loop(self):
        """Lance la boucle asyncio dans un thread séparé."""
        set_event_loop(self.__loop)
        self.__loop.run_forever()

    def start(self) -> None:
        """Démarre le bot"""

        self.__running_bot: Future = run_coroutine_threadsafe(self.__bot.start(token=self.__token, reconnect=True), self.__loop)

    def stop(self) -> None:
        """
        Stop le bot proprement depuis un autre thread
        :raise: BotNotStartedError
        """
        if self.is_running:
            # Attendre que la fermeture du bot soit terminée
            run_coroutine_threadsafe(self.__stop_bot_in_thread(), self.__loop).result(timeout=30)
            self.__bot = Bot(token=self.__token, intents=self.__intents, command_prefix=self.__command_prefix, help_command=None)
            self.__running_bot = None
        else:
            raise BotNotStartedError(self.__bot.user.name)

    async def __stop_bot_in_thread(self):
        """
        Clear le cache du bot de manière asynchrone
        """
        await self.__bot.close()

    def add_pyFile_commands(self, file: str, reload_command: bool, setup_function: str):
        """
        Ajoute et charge un fichier de commande bot et ses dépendances.
        Les fichiers doivent avoir une fonction appelée « setup » ou un équivalent passé en paramètre.

        def setup(bot: Bot) :
            ...

        :param bot_name : Le nom du bot à ajouter au fichier de commandes
        :param file: Chemin relatif ou absolue du fichier de commande
        :param setup_function : Nom de la fonction appelée par le processus pour donner l'instance de Bot.
        :param reload_command : Recharge toutes les commandes dans le fichier et les dépendances. Défaut : True
        """
        module_name = path.splitext(path.basename(file))[0] # récupère le nom du fichier
        spec = spec_from_file_location(module_name, file) # renvoie un "module spec" à partir du nom et du fichier
        if spec and spec.loader:
            module = module_from_spec(spec) # crée le package à partir du "module spec" s'il existe
            spec.loader.exec_module(module) # charge tout le package
            modules[module_name] = module # enregistre le modul dans les packages du système pour qu'il soit retrouvable lors du rechargement de celui-ci
            self.__call_setup_function_in_command_file(file, module, setup_function, reload_command)
        else:
            raise CommandFileNotFoundError(file)

    def modify_pyFile_commands(self, file: str, setup_function: str):
        """
        Modifie un fichier de commandes et le recharge.
        Ne recharge que le fichier et non les commandes du bot !
        :param file: Le chemin d'accès relatif ou absolue du fichier
        """
        module_name = path.splitext(path.basename(file))[0]
        module_found = False

        # Mise à jour du module et de son setup
        for imported in self.__imported_module:

            if imported['module'].__name__ == module_name:
                imported['setup_function'] = setup_function
                module_found = True
                break

        if not module_found:
            raise CommandFileNotFoundError(file)

        # Supprimer toutes les commandes du bot
        for command in self.__bot.application_commands:
            self.__bot.remove_application_command(command)

        # Réattacher toutes les commandes en réexécutant tous les setup
        for imported in self.__imported_module:
            self.__call_setup_function_in_command_file(
                file,
                imported['module'],
                imported['setup_function'],
                reload_command=False
            )

    def reload_pyFile_commands(self):
        """
        Recharge tous les fichiers de commandes du bot
        """
        for imported in self.__imported_module:
            self.modify_pyFile_commands(imported['file'], imported['setup_function'])



    def __call_setup_function_in_command_file(self, file: str, module, setup_function: str, reload_command: bool):
        """
        Appel la fonction de setup du module pour charger toutes les commandes du bot
        :param file: Le chemin d'accès du fichier de commandes
        :param module: Le module préchargé
        :param setup_function: Le nom de la fonction de setup
        :param reload_command: Si les commandes doivent être recharger sur le bot (envoie une requête à Discord) automatiquement
        """
        if hasattr(module, setup_function):  # si la fonction setup (ou autre) est dans le package
            getattr(module, setup_function)(self.__bot)

            ########## permet de modifier le dictionnaire des modules importés si celui-ci existe déjà, sinon il l'ajoute à la liste des dictionaires des modules importés. Utile car on reload tout si un des modules est modifié
            find = False
            for mod in self.__imported_module:
                if mod['module'].__name__ == module.__name__:
                    mod['setup_function'] = setup_function
                    find = True
                    break

            if not find:
                self.__imported_module.append({'setup_function': setup_function, 'module': module, 'file': file})
            ##########

            if reload_command:
                self.reload_commands()
        else:
            raise SetupCommandFunctionNotFound(setup_function, file)

    def reload_commands(self, commands: Optional[list[ApplicationCommand]] = None):
        """
        Charge toutes les commandes du bot sur Discord
        """
        run_coroutine_threadsafe(self.__reload_commands(commands=commands), self.__loop).result(timeout=60)

    async def __reload_commands(self, commands: Optional[list[ApplicationCommand]] = None):
        """
        Recharge les commandes quand le bot est ready
        """
        if self.__running_bot is not None:
            while not self.is_ready:
                await sleep(0.3)
            await self.__bot.register_commands(commands=commands, method='individual', force=False)
        else:
            raise BotNotStartedError(self.__token)


    @property
    def is_running(self) -> bool:
        """Renvoie si la Websocket est connectée"""
        return not self.__bot.is_closed()

    @property
    def is_ready(self) -> bool:
        """
        Renvoie si le bot est ready
        """
        return self.__bot.is_ready()

    @property
    def is_ws_ratelimited(self) -> bool:
        """
        Renvoie si le bot est rate limit
        """
        return self.__bot.is_ws_ratelimited()


    def close_ascyncio_loop(self):
        """
        Ferme la boucle asyncio
        """
        if self.__loop.is_running():
            self.__loop.stop()

        while self.__loop.is_running():
            tsleep(0.3)

        self.__loop.close()





from ..views.easy_modified_view import EasyModifiedViews

from typing import Optional, Union, Any, Callable
from discord.ui import Button
from discord import ButtonStyle, Role, Interaction
from inspect import currentframe

class Back:
    """
    Class
    -------------
    Allows you to easily setup a back button.
    He save function parameters when he is called and recall the function with these parameters.
    -------------
    """

    def __init__(self, timeout: Optional[float] = None,
                 disabled_on_timeout: bool = False,
                 row: int = 0,
                 autorised_roles: Optional[list[Union[Role, int]]] = None,
                 autorised_key: Optional[callable] = None):
        """
        Initialisation for back button
        """
        self.args_function: tuple = ()
        self.kwargs_function: dict = {}

        # get the calling function
        function_frame = currentframe().f_back
        call_function_frame = function_frame.f_back
        name = function_frame.f_code.co_name
        self.back_function: Union[Callable] = call_function_frame.f_globals.get(name) or call_function_frame.f_locals.get(name)

        self.__view = EasyModifiedViews(timeout=timeout, disabled_on_timeout=disabled_on_timeout)
        self.__view.add_items(Button(label='⏪ Back', row=row, custom_id='back', style=ButtonStyle.gray))
        self.__view.set_callable('back', _callable=self._back, autorised_roles=autorised_roles, autorised_key=autorised_key)

    def set_parameters(self, *args, **kwargs):
        """
        Set the parameters to recall the function
        :param args: The args to save
        :param kwargs: The kwargs to save
        """
        self.args_function = args
        self.kwargs_function = kwargs

    async def _back(self, button: Button, interaction: Interaction, data: dict[str, Any]):
        """
        Base asynchronous _back function called when button is pressed
        """
        await interaction.response.defer()
        await self.back_function(*self.args_function, **self.kwargs_function)

    def get_view(self) -> EasyModifiedViews:
        """
        Get the view with the back button
        :return: The view with the back button
        """
        return self.__view


from __future__ import annotations
from typing import Optional, Callable
from discord import File, Embed
from immutableType import callable_, NoneType
from ..views import EasyModifiedViews

class Page:

    @callable_(is_class=True, kwargs_types={'view': [NoneType, EasyModifiedViews], 'content': [NoneType, str], 'embed': [NoneType, Embed], 'embeds': [list], 'file': [NoneType, File], 'files': [list]})
    def __init__(self, view: EasyModifiedViews, content: Optional[str] = None, embed: Optional[Embed] = None, embeds: list[Embed] = [], file: Optional[File] = None,  files: Optional[list[File]] = []):
        """
        Init Page instance from Pagination class
        """
        self.__view: EasyModifiedViews = view if view is not None else EasyModifiedViews()
        self.content: Optional[str] = content
        self.embeds: list[Embed] = embeds if embed is None else embeds + [embed]
        self.files: list[File] = files if file is None else files + [file]

    @property
    def get_page_view(self) -> EasyModifiedViews:
        """
        Get the current page view
        """
        return self.__view

